use super::super::types::{BootOverlaySection, OverlayDesc, OverlaySection, OverlaySlot};
use tool_helper::{Error, format_if_error, Output};
use std::io::Write;

const OVERLAY_DEFAULT_START:&'static str = "__planschi_start"; // < will probably be "__boot_loader_start" later

pub fn write(output: &mut Output, overlay_desc: &OverlayDesc) -> Result<(), Error> {
    fn create_heap_base(overlay_desc: &OverlayDesc) -> String {
        let (first_entry,additional_slots_iter) = {
            if overlay_desc.boot_section.is_empty() {
                (&overlay_desc.additional_slots[0].name, overlay_desc.additional_slots.iter().skip(1))
            }

            else {
                (&overlay_desc.boot_section.name, overlay_desc.additional_slots.iter().skip(0))
            }
        };

        let mut heap_start = format!("MAX({}, __{}_end)", OVERLAY_DEFAULT_START, first_entry);
        for slot in additional_slots_iter {
            heap_start = format!("MAX(__{}_end, {})", slot.name, heap_start);
        }
        heap_start
    }

    fn get_slot_start_adr(slot: &OverlaySlot) -> &str {
        slot.start_adr.as_ref()
    }

    if overlay_desc.additional_slots.is_empty() && overlay_desc.boot_section.is_empty() {
        write_heap_base(output, OVERLAY_DEFAULT_START)?;
    }

    else {
        write_heap_base(output, &create_heap_base(&overlay_desc))?;
    }

    writeln!(output, "")?;
    writeln!(output, "SECTIONS {{")?;
        writeln!(output, "\t/DISCARD/ : {{ *(.comment*) *(.gnu.*) *(.mdebug*) *(.debug_*) *(.pdr)}}")?;
        write_first_section(output, &overlay_desc.boot_section)?;
        for slot in &overlay_desc.additional_slots {
            writeln!(output, "\t/*{}*/", slot.name)?;
            writeln!(output, "\tOVERLAY {} : NOCROSSREFS SUBALIGN(4) {{", get_slot_start_adr(slot))?;
                write_section(output, &slot.sections, false)?;
            writeln!(output, "\t}}")?;
            writeln!(output, "\t. = ALIGN(4);")?;
            writeln!(output, "\t__{}_end = .;", slot.name)?;
            writeln!(output, "")?;
        }
    writeln!(output, "}}")?;
    return Ok(());
}

fn write_heap_base(output: &mut Output, heap_base: &str) -> Result<(), Error> {
    format_if_error!(writeln!(output, "__heap_base = {};", heap_base), "Writing default LD Script failed with: {error_text}")?;
    Ok(())
}

fn write_first_section(output: &mut Output, boot_overlay: &BootOverlaySection) -> Result<(), Error> {
    writeln!(output, "\tOVERLAY __bss_end : NOCROSSREFS SUBALIGN(4) {{")?;
    writeln!(output, "\t\t/*{}*/", boot_overlay.name)?;
    write_section(output, &boot_overlay.sections, true)?;
    writeln!(output, "\t}}")?;
    writeln!(output, "\t. = ALIGN(4);")?;
    writeln!(output, "\t__{}_end = .;", boot_overlay.name)?;
    writeln!(output, "")?;

    Ok(())
}

fn write_section(output: &mut Output, sections: &Vec<OverlaySection>, add_end_name: bool) -> Result<(), Error> {
    for section in sections {
        writeln!(output, "\t\t.{} {{", section.name)?;
            writeln!(output, "\t\t\t__{}_start = .;", section.name)?;
            section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\tKEEP({}(.header))", patr))?;
            writeln!(output, "\t\t\t__{}_lbas = .;", section.name)?;
            section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\tKEEP({}(.header.lbas))", patr))?;
            writeln!(output, "\t\t\t__{}_ctor = .;", section.name)?;

            for section_type in [".text.startup._GLOBAL__*", ".ctors", ".text.*", ".rodata*", ".sdata*", ".data*", ".sbss*", ".bss*"] {
                section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\t{}({})", patr, section_type))?;
            }

            if add_end_name {
                writeln!(output, "\t\t\t__{}_end = .;", section.name)?;
            }
        writeln!(output, "\t\t}}")?;
    }

    Ok(())
}